"								 Version 1.2.1
" format.vim -	Format multibyte text, for tha languages, which can split
"		line anywhere, unless prohibited.
"
" Written by:	  Taro Muraoka <koron@tka.att.ne.jp>
" Modified by:	  Takuhiro Nishioka <takuhiro@super.win.ne.jp>
" Last change:	  2000 Jun 6
"
" function Format(start_line_number, end_line_number)
" 
" Format() will allow format multibyte text.  In some of East Asian
" languages, the line can break anywhere, unless prohibited.  Original Vim's
" "gq" format command doesn't allow to break line at the midst of word.
" This function split line at each multibyte character.  And it can handle
" prohibited line break rules.
"
" This function is following Vim's "gq" command. But there will be lack of
" something.


"---------------------------------------------------------------------------
"				    Options
"
" "$format_command"
"
" Specifies the format command that format lines to the width the
" 'textwidth' option specifies. The "Q" command formerly did this, so if you
" still want to use "Q", set this to "Q"
"
"let $format_command = "Q"
let $format_command = "gq"

"
" "$format_join_spaces"
"
" Delete a space , when joining lines, according to the following rules.
" This rule is also applied to "J" command
"   1	line end with AND next line start with a multibyte char
"   2	line end with OR  next line start with a multibyte char
"   3	same with original join command
"
let $format_join_spaces = 1

"
" "$format_follow_taboo_rule"
"
" Move to a point that will not break forbidden line break rules. If you
" don't want to do this, set this to "0".
"
let $format_follow_taboo_rule = 1

"
" "$format_allow_over_tw"
"
" The width that can over 'textwidth'. This variable is used for taboo rule.
"
let $format_allow_over_tw = 2

"
" "$format_no_begin"
"
" This option is space-separated list of characters, that are forbidden to
" be at beginning of line. Add two spaces for ASCII characters. See also
" TabooRuleMatch()
"
let $format_no_begin = "!  ,  .  ?  )  ]  }  -  _  ~  :  ;  "
if &fileencoding == "japan"
  let $format_no_begin = $format_no_begin
    \. "  f h     A B X r t v x z l "
    \. "          J K T U "
    \. "@ B D F H b       "
    \. "E [ R S I  j C D F G H n p c ` "
endif

"
" "$format_no_end"
"
" This option is space-separated list of characters, that are forbidden to
" be at end of line. Add two spaces for ASCII characters. See also
" TabooRuleMatch()
"
let $format_no_end = "(  [  {  "
if &fileencoding == "japan"
  let $format_no_end = $format_no_end
    \. " e g q s u w y k  i m o  "
endif

"
" "$format_motion_list"
"
" Specifiles the motion command that could follow "$format_command". These
" are pairs of two letters. Note that not all motion commands can be
" specified here. Use visual mode for other motions.
"
let $format_motion_list = "apawaWasa]a[a)a(aba>a}a{aB"
  \. "ipiwiWisi]i[i)i(ibi>i}i{iB"
  \. "gqq "
  \. "h j k l 0 ^ $ ; , - + w W e E b B ( ) { } n N % H M L G / ? "
  \. "gjgkggg0g^gmgegEg$]][[][[[[]''``'[`[']`]'<`<'>`>'\"`\""
  \. "[([{])]}]m]M[m[M[#]#[*[/"
  \. "'a'b'c'd'e'f'g'h'i'j'k'l'm'n'o'p'q'r's't'u'v'w'x'y'z"
  \. "`a`b`c`d`e`f`g`h`i`j`k`l`m`n`o`p`q`r`s`t`u`v`w`x`y`z"

"
" "$format_indent_sensitive"
"
" When the indentation changes, it's the end of a paragraph. Note that if
" this option is set, second indentation is disabled.
"
let $format_indent_sensitive = 0

"---------------------------------------------------------------------------
"
" DoMappings()
"   Do mappings.
"
function! DoMappings()
  " Normal mode mappings.
  let length = strlen($format_motion_list)
  let i = 0
  while i <= length - 2
    let motion = strpart($format_motion_list, i, 2)
    let motion = substitute(motion, " $", "", "")
    execute "nmap " . $format_command . motion . " :call "
      \. "FormatWorkhorse(\"" . motion . "\")<CR>"
    let i = i + 2
  endwhile

  " Visual mode mapping.
  execute "vmap " . $format_command . " <ESC>:call"
    \. " Format(line(\"'<\"), line(\"'>\"))<CR>"

  " Change "J" to follow "$format_join_spaces"
  nmap J :call DoJoin("")<CR>
  vmap J <ESC>:call DoJoinRange(line("'<"), line("'>"))<CR>
endfunction

"
" FormatWorkhorse(motion)
"   Select the area that moves over, then pass the start and end line number
"   of the area to Format()
"
function! FormatWorkhorse(motion)
  if a:motion == "gq" || a:motion == "q"
    execute "normal! V\<ESC>"
  elseif a:motion == "/" || a:motion == "?"
    let pattern = input("Enter pattern: ")
    execute "normal! v" . a:motion . pattern . "\<CR>\<ESC>"
  else
    execute "normal! v" . a:motion . "\<ESC>"
  endif
  call Format(line("'<"), line("'>"))
endfunction

"
" Format(start_lnum, end_lnum)
"   Format the area from the start line number to the end line number.
"
function! Format(start_lnum, end_lnum)
  let count_nr = a:end_lnum - a:start_lnum + 1
  let advance = 1
  " current line is the start of a paragraph.
  let first_par_line = 1
  " the second indent
  let second_indent = "default"

  " Check 2 in the formatoptions
  let do_second_indent = HasFormatOptions('2')

  let wrap_save = &wrap
  set nowrap
  let showcmd_save = &showcmd
  set noshowcmd

  " Set cursor to the start line number.
  call SetCorsor(a:start_lnum)

  " Get info about the previous and current line.
  if a:start_lnum == 1
    " current line is not part of paragraph
    let is_not_par = 1
  else
    normal! k
    " the commet leader of current line
    let leader = GetLeader()
    let is_not_par = FmtCheckPar(leader)
    normal! j
  endif

  " the commet leader of next line
  let next_leader = GetLeader()
  " next line not part of paragraph
  let next_is_not_par = FmtCheckPar(next_leader)

  " at end of paragraph
  let is_end_par = is_not_par || next_is_not_par

  " operation top
  let op_top = 1
  while count_nr > 0
    " Advance to next paragraph.
    if advance
      if op_top
	let op_top = 0
      else
	normal! j
      endif
      let leader = next_leader
      let is_not_par = next_is_not_par
      " previous line is end of paragraph
      let prev_is_end_par = is_end_par
    endif

    " The last line to be formatted.
    if count_nr == 1
      let next_leader = ""
      let next_is_not_par = 1
    else
      normal! j
      let next_leader = GetLeader()
      let next_is_not_par = FmtCheckPar(next_leader)
      normal! k
    endif

    let advance = 1
    let is_end_par = is_not_par || next_is_not_par

    " Skip lines that are not in a paragraph.
    if !is_not_par

      " For the first line of a paragraph, check indent of second line.
      " Don't do this for comments and empty lines.
      if first_par_line
	  \&& do_second_indent
	  \&& prev_is_end_par
	  \&& leader =~ "^\\s*$"
	  \&& next_leader =~ "^\\s*$"
	  \&& getline(line(".") + 1) !~ "^$"
	let second_indent = next_leader
      endif

      " When the comment leader changes, it's the end of the paragraph
      if !SameLeader(leader, next_leader)
	let is_end_par = 1
      endif

      " If we have got to the end of a paragraph, format it.
      if is_end_par
	" do the formatting
	call FormatLine(second_indent)
	let second_indent = "default"
	let first_par_line = 1
      endif

      " When still in same paragraph, join the lines together.
      if !is_end_par
	let advance = 0
	" join current line and next line without the comment leader
	call DoJoin(next_leader)
	let first_par_line = 0
      endif

    endif
    let count_nr = count_nr - 1
  endwhile
  if wrap_save
    set wrap
  endif
  if showcmd_save
    set showcmd
  endif
endfunction

"
" FormatLine(second_indent)
"   Format currentline.
"
function! FormatLine(second_indent)
  " check textwidth
  if &textwidth == 0
    let textwidth = 76
  else
    let textwidth = &textwidth
  endif

  let do_second_indent = HasFormatOptions("2")
  let fo_do_comments = HasFormatOptions("q")
  let second_indent = a:second_indent

  " save the original option's value
  let formatoptions_save = &formatoptions
  let iskeyword_save = &iskeyword

  let leader_width = GetLeader("get_leader_width")

  " When fo_do_comments is TRUE, set formatoptions value so that the comment
  " leader is set for next line.
  if fo_do_comments
    set formatoptions+=r
  else
    set formatoptions-=r
  endif

  " Set iskeyword option value to every printable ascii characters, so that
  " "w" can stop at only multibyte-ascii boundary or white space.
  set iskeyword="!-~"

  call SetCorsor(line("."), textwidth)
  while GetWidth() > virtcol(".")
    let finish_format = 0
    let force_fold = 0
    let do_insert = 0
    let max_width = virtcol(".") + $format_allow_over_tw

    let ch = GetCharUnderCursor()
    normal! l
    let next_ch = GetCharUnderCursor()
    normal! h

    " English word folding
    if ch =~ "[!-~]\\{1}" && next_ch =~ "[!-~]\\{1}"
      call MoveToWordBegin()
      if virtcol(".") - 1 > leader_width
	" move to previous word end
	normal! ge
      endif
    endif

    " Skip white spaces
    if ch =~ "\\s"
      while ch =~ "\\s" && virtcol(".") - 1 > leader_width
	normal! h
	let ch = GetCharUnderCursor()
      endwhile
      let force_fold = 1
    endif

    if virtcol(".") - 1 <= leader_width
      call MoveToFirstWordEnd(leader_width)
      let force_fold = 1
      if GetWidth() == virtcol(".")
	let finish_format = 1
      endif
    endif

    " Taboo rule
    if !finish_format && !force_fold && $format_follow_taboo_rule
      normal! l
      let next_ch = GetCharUnderCursor()
      normal! h
      if TabooRuleMatch($format_no_begin, next_ch)
	normal! l
	while TabooRuleMatch($format_no_begin, next_ch)
	  " if cursor is at the line end, break.
	  if GetWidth() == virtcol(".")
	    let finish_format = 1
	    break
	  endif
	  normal! l
	  let next_ch = GetCharUnderCursor()
	endwhile
	if !finish_format
	  normal! h
	endif
      endif

      let ch = GetCharUnderCursor()
      if virtcol(".") > max_width
	let finish_format = 0
	while TabooRuleMatch($format_no_begin, ch)
	    \&& virtcol(".") - 1 > leader_width
	  normal! h
	  let ch = GetCharUnderCursor()
	endwhile
	if ch =~ "[!-~]\\{1}"
	  call MoveToWordBegin()
	  if virtcol(".") - 1 > leader_width
	    normal! ge
	  else
	    call MoveToFirstWordEnd(leader_width)
	    let force_fold = 1
	  endif
	else
	  let do_insert = 1
	endif
      endif

      let ch = GetCharUnderCursor()
      if TabooRuleMatch($format_no_end, ch) && !force_fold
	let do_insert = 0
	while TabooRuleMatch($format_no_end, ch)
	    \&& virtcol(".") -1 > leader_width
	  normal! h
	  let ch = GetCharUnderCursor()
	endwhile
	if virtcol(".") -1 <= leader_width
	  call MoveToFirstWordEnd(leader_width)
	endif
      endif
    endif

    if finish_format
      break
    endif

    if do_insert
      call InsertNewLine()
    else
      call AppendNewLine()
    endif

    if do_second_indent && second_indent != "default"
      call setline(line(".")
	  \, second_indent . substitute(getline("."), "^\\s*", "", ""))
      let do_second_indent = 0
      if strlen(second_indent) > 0
	normal! h
      endif
    endif

    if virtcol(".") == 1
      let leader_width = 0
    else
      let leader_width = virtcol(".")
    endif

    call SetCorsor(line("."), textwidth)
  endwhile

  execute "set formatoptions=" . formatoptions_save
  execute "set iskeyword=" . iskeyword_save
endfunction

"
" GetLeader(...)
"   Get the comment leader string from current line. If argument
"   is specified, then return the comment leader width. Note that
"   returned comment leader and the current line's comment leader is
"   not always same.
"
function! GetLeader(...)
  if !HasFormatOptions('q')
    if a:0 == 1
      return 0
    endif
    return ""
  endif

  let col_save = virtcol(".")

  let formatoptions_save = &formatoptions
  let autoindent_save = &autoindent
  let cindent_save = &cindent
  let smartindent_save = &smartindent
  set formatoptions+=o
  set autoindent
  set nocindent
  set nosmartindent

  execute "normal! ox\<ESC>\"_x"

  if a:0 == 1
    if getline(".") =~ "^$"
      let leader_width = 0
    else
      let leader_width = virtcol(".")
    endif
  endif

  let leader = getline(".")

  if line(".") == line("$")
    normal! "_dd
  else
    normal! "_ddk
  endif

  execute "set formatoptions=" . formatoptions_save
  if !autoindent_save
    set noautoindent
  endif
  if cindent_save
    set cindent
  endif
  if smartindent_save
    set smartindent
  endif

  execute "normal! " . col_save . "|"

  if a:0 == 1
    return leader_width
  else
    return leader
  endif
endfunction

"
" FmtCheckPar(leader)
"   Blank lines, lines containing only white space or the comment leader,
"   are left untouched by the formatting. The function returns true in this
"   case.
"
function! FmtCheckPar(leader)
  let three_start = substitute(&com, '.*s[^:]*:\([^,]*\),.*', '\1', '')
  let three_end = substitute(&com, '.*e[^:]*:\([^,]*\),.*', '\1', '')
  let line = substitute(getline("."), "\\s*$", "", "")
  let line = substitute(line, "^\\s*", "", "")
  let leader = substitute(a:leader, "\\s*$", "", "")
  let leader = substitute(leader, "^\\s*", "", "")
  if line == three_start || line == three_end
    return 1
  endif
  return line == leader
endfunction

"
" SameLeader(leader1, leader2)
"   Return true if the two comment leaders given are the same. White-space is
"   ignored.
"
function! SameLeader(leader1, leader2)
  if $format_indent_sensitive
    return a:leader1 == a:leader2
  else
    return substitute(a:leader1, "\\s\\+$", "", "")
	\== substitute(a:leader2, "\\s\\+$", "", "")
  endif
endfunction


"
" SetCorsor(lnum, width)
"   Set corsor to the line number, then move the cursor to within the width
"   and the most right virtual column.
"
function! SetCorsor(lnum, ...)
  execute a:lnum

  if a:0 == 1
    let wrap_save = &wrap
    set nowrap
    execute "normal! " . a:1 . "|"
    if virtcol(".") == a:1
	\&& strlen(GetCharUnderCursor()) > 1
	\&& virtcol(".") > 1
      normal! h
    endif
    if wrap_save
      set wrap
    endif
  endif
endfunction

"
" HasFormatOptions(x)
"   Return true if format option 'x' is in effect. Take care of no
"   formatting when 'paste' is set.
"
function! HasFormatOptions(x)
  if &paste
      \|| (a:x == "2" && !&autoindent)
      \|| (a:x == "2" && $format_indent_sensitive)
    return 0
  endif
  return &formatoptions =~ a:x
endfunction

"
" DoJoin(next_leader)
"   Join line and next line ,according to $format_join_spaces. The comment
"   leader will be removed.
"
function! DoJoin(next_leader)
  if line(".") == line("$")
    return
  endif

  let wrap_save = &wrap
  set nowrap
  let showcmd_save = &showcmd
  set noshowcmd

  normal! $
  let end_char = GetCharUnderCursor()

  if HasFormatOptions("q") && a:next_leader != ""
    let next_leader = escape(a:next_leader, '^.*\$~[]')
    let next_leader = "^" . substitute(next_leader, "\\s*$", "", "")
    normal! j0
    if getline(".") =~ next_leader
      call setline(line("."), substitute(getline("."), next_leader, "", ""))
    else
      let leader_width = GetLeader("get_leader_width")
      let i = leader_width + 1
      execute "normal! 0\"_d" . i . "|"
    endif
    normal! k
  endif

  normal! J

  if GetWidth() > virtcol(".") && GetCharUnderCursor() == " "
    normal! l
    let begin_char = GetCharUnderCursor()
    normal! h
    if $format_join_spaces == 1
	\&& (strlen(end_char) > 1 && strlen(begin_char) > 1)
      normal! "_x
    elseif $format_join_spaces == 2
	\&& (strlen(end_char) > 1 || strlen(begin_char) > 1)
      normal! "_x
    endif
  endif
  if wrap_save
    set wrap
  endif
  if showcmd_save
    set showcmd
  endif
endfunction

function! DoJoinRange(start_lnum, end_lnum)
  let count_nr = a:end_lnum - a:start_lnum
  call SetCorsor(a:start_lnum)
  while count_nr > 0
    call DoJoin("")
    let count_nr = count_nr - 1
  endwhile
endfunction

"
" GetWidth()
"   Return the current line width. If the line is empty returns 0. Note that
"   if the character at the line end is a multibyte character, this returns
"   real width minus 1, same as virtcol().
"
function! GetWidth()
  let col_save = virtcol(".")
  normal! $
  let width = virtcol(".")
  execute "normal! " . col_save . "|"
  return width
endfunction

"
" GetCharUnderCursor()
"   Get (multibyte) character under current cursor.
"
function! GetCharUnderCursor()
  let str = getline(".")
  let idx = col(".") - 1
  let ch = str[idx]
  if char2nr(ch) >= 128
    return strpart(str, idx, 2)
  else
    return ch
  endif
endfunction

"
" AppendNewLine()
"   Insert newline after cursor.
"
function! AppendNewLine()
  execute "normal! a\<CR>\<ESC>"
endfunction

"
" InsertNewLine()
"   Insert newline before cursor.
"
function! InsertNewLine()
  execute "normal! i\<CR>\<ESC>"
endfunction

"
" MoveToWordEnd()
"   Move to the word end.
"
function! MoveToWordEnd()
  if line(".") == 1
    normal! wge
  else
    normal! gee
  endif
endfunction

"
" MoveToWordBegin()
"   Move to the word begin.
"
function! MoveToWordBegin()
  if line(".") == 1
    normal! wb
  else
    normal! gew
  endif
endfunction

"
" MoveToFirstWordEnd()
"   Move to the first word end after the comment leader.
"
function! MoveToFirstWordEnd(leader_width)
  let i = a:leader_width + 1
  execute "normal! " . i . "|"
  call MoveToWordEnd()
endfunction

"
" TabooRuleMatch(taboo_rule_list, char)
"   Return true when the character matches one of taboo_rule_list
"
function! TabooRuleMatch(taboo_rule_list, char)
  " add spaces to char so to match exactly one of the list
  if strlen(a:char) > 1
    let ch = a:char . " "
  else
    let ch = a:char . "  "
  endif

  " escape the special character
  return a:taboo_rule_list =~ escape(ch, '^.*\$~[]')
endfunction

call DoMappings()

" vi:ts=8 sts=2 sw=2 tw=0 com=\:\"
